#ifdef __cplusplus
extern "C" 
{
#endif
#include <baby_plugin.h>
#include <png.h>

void _init();
void _fini();
//-----the following is the callback for main programe
//-----NOTE:i defined them static,because there's no need to call outside,i just provide the function pointer to main programe.
static int i_support(const char*file,const char*ext,UCHAR*magic,UINT len,int option);
static int load_preview(const char*file,BabyImage*bb,UINT dw,UINT dh,int option);
static int load_image(const char*file,BabyImage*bb,int option);
static int free_image(BabyImage*bb);
static int i_can_save(const char*file,const char*saveext);
static int save_image(const char*file,const char*saveext,const BabyImage*bb,const BabyOptions*bbs);

static GetMemory	get_memory=NULL;
static FreeMemory	free_memory=NULL;
static LockX		lock_x=NULL;
static UnlockX		unlock_x=NULL;

static const char	*my_open[]={"PNG Image (*.png)",NULL};
static const char	*my_save[]={"PNG Image (*.png)",NULL};
//--------------------------------------------------------------------
int
BabyBorn(BabyImageEntry*bb)
{
	get_memory=bb->utils.get_memory;
	free_memory=bb->utils.free_memory;
	lock_x=bb->utils.lock_x;
	unlock_x=bb->utils.unlock_x;

	if(bb->utils.language&&strcasecmp(bb->utils.language,"gbk")==0)
	{
		bb->info.plug_name="PNG";
		bb->info.discription="BabyseePNG֧PNGʽļغͱ档";
		bb->info.author="walk_in_sky  falconinsky";
		bb->info.copyright="";
	}
	else
	{
		bb->info.plug_name="PNG Plugin";
		bb->info.discription="Babysee PNG plugin, support png load,save.";
		bb->info.author="walk_in_sky & falconinsky.";
		bb->info.copyright="Freeware.";
	}
	bb->info.open_information=my_open;
	bb->info.save_information=my_save;
	
	bb->open_s.do_support=i_support;
	bb->open_s.load_preview=load_preview;
	bb->open_s.load_image=load_image;
	bb->open_s.free_image=free_image;
	
	bb->save_s.can_you_save=i_can_save;
	bb->save_s.save_image=save_image;
	
	return (get_memory&&free_memory)?1:0;//if you want to use Xlib,please check lock_x and unlock_x
}
//-----------------------------------------------------------------------------------------------------
void _init()
{
	
}
void _fini()
{
	
}
int	
i_support(const char*file,const char*ext,UCHAR*magic,UINT len,int option)
{
	if(file==NULL)return 0;
	if(ext==NULL)
	{
		if(magic==NULL||len<8)
		{
			UCHAR	buf[9];
			int 	iret;
			int		fp=open(file,O_RDONLY);
			if(fp<=0)return 0;
			iret=read(fp,buf,8);
			close(fp);
			return ((iret==8)&&png_check_sig(buf, 8))?1:0;
		}
		return (png_check_sig(magic, 8))?1:0;
	}
	if(strcasecmp(ext,"png")==0)return 1;
	return 0;
}
int
load_preview(const char*file,BabyImage*bb,UINT dw,UINT dh,int option)
{
	png_structp			png_ptr;
	png_infop			info_ptr;
	UCHAR				*data, *ptr, **lines, *ptr2, r, g, b, a;
	int					transp, bit_depth, color_type, interlace_type;
	png_uint_32			ww, hh;
	FILE				*f=NULL;
	UINT				i,x,y;
  /* Init PNG Reader */
	transp = 0;
	data=NULL;
	lines=NULL;
	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr)return 0;
	info_ptr = png_create_info_struct(png_ptr);
	if (!info_ptr)
	{/**/
		png_destroy_read_struct(&png_ptr, NULL, NULL);
		return 0;
	}
	if (setjmp(png_ptr->jmpbuf))
	{/**/
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return 0;
	}
	if (info_ptr->color_type == PNG_COLOR_TYPE_RGB_ALPHA)
	{/**/
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		return 0;
	}
	f=fopen(file,"rb");
	if(f==NULL)
	{
		png_destroy_read_struct(&png_ptr,&info_ptr,NULL);
		return 0;
	}
	png_init_io(png_ptr, f);

	/* Read Header */
	png_read_info(png_ptr, info_ptr);
	png_get_IHDR(png_ptr, info_ptr, &ww, &hh, &bit_depth, &color_type, &interlace_type, NULL, NULL);
	bb->num_images=1;
	bb->image=(BabyOneImage*)get_memory(sizeof(BabyOneImage));
	bzero(bb->image,sizeof(BabyOneImage));
	bb->image->image_width=ww;
	bb->image->image_height=hh;
	if(bb->image->image_width>32767||bb->image->image_height>32767)
	{
		png_destroy_read_struct(&png_ptr,&info_ptr,NULL);
		fclose(f);
		free_memory(bb->image);
		return 0;
	}
	bb->image->rgb_width=ww;
	bb->image->rgb_height=hh;
	bb->image->image_bpp=1<<bit_depth;
	
	/* Setup Translators */
	if ((color_type == PNG_COLOR_TYPE_PALETTE) || (color_type == PNG_COLOR_TYPE_GRAY)) png_set_expand(png_ptr);
	png_set_strip_16(png_ptr);
	png_set_packing(png_ptr);
	if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))  png_set_expand(png_ptr);

	png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);
	data =(UCHAR*)get_memory(ww*hh*3);
	lines = (UCHAR**)malloc(hh * sizeof(UCHAR*));
	if (!data || !lines)
	{/**/
		if(data)free_memory(data);
		if(lines)free(lines);
		png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
		fclose(f);
		free_memory(bb->image);
		return 0;
	}
	bb->image->rgb_data=data;

	for (i = 0; i < hh; i++)
	{/**/
		if ((lines[i] = (UCHAR*)malloc(ww * 4)) == NULL)
		{/**/
			UINT			n;
			free_memory(data);
			for (n = 0; n < i; n++)
				free(lines[n]);
			free(lines);
			png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
			fclose(f);
			free_memory(bb->image);
			return 0;
		}
	}

	png_read_image(png_ptr, lines);
	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
	fclose(f);
	ptr = data;
	
	if (color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
	{/**/
		for (y = 0; y < hh; y++)
		{/**/
			ptr2 = lines[y];
			for (x = 0; x < ww; x++)
			{/**/
				r = *ptr2++;
				a = *ptr2++;
				if (a < 128)
				{/**/
					*ptr++ = 255;
					*ptr++ = 0;
					*ptr++ = 255;
					transp = 1;
				}
				else
				{/**/
					*ptr++ = r;
					*ptr++ = r;
					*ptr++ = r;
				}
			}
		}
	}
	else if (color_type == PNG_COLOR_TYPE_GRAY)
	{/**/
		for (y = 0; y < hh; y++)
		{/**/
			ptr2 = lines[y];
			for (x = 0; x < ww; x++)
			{/**/
				r = *ptr2++;
				ptr2++;
				*ptr++ = r;
				*ptr++ = r;
				*ptr++ = r;
			}
		}
	}
	else
	{/**/
		for (y = 0; y < hh; y++)
		{/**/
			ptr2 = lines[y];
			for (x = 0; x < ww; x++)
			{/**/
				r = *ptr2++;
				g = *ptr2++;
				b = *ptr2++;
				a = *ptr2++;
				if (a < 128)
				{/**/
					*ptr++ = 255;
					*ptr++ = 0;
					*ptr++ = 255;
					transp = 1;
				}
				else
				{/**/
					if ((r == 255) && (g == 0) && (b == 255))
						r = 254;
					*ptr++ = r;
					*ptr++ = g;
					*ptr++ = b;
				}
			}
		}
	}
	for (i = 0; i < hh; i++)
		free(lines[i]);
	free(lines);
	if(transp)
	{
		bb->image->mask_color.r=255;
		bb->image->mask_color.g=0;
		bb->image->mask_color.b=255;		
	}
	return 1;
}
int
load_image(const char*file,BabyImage*bb,int option)
{
	return load_preview(file,bb,0,0,option);
}


int
free_image(BabyImage*bb)
{
	if(!bb)return 0;
	return 1;
}

int 
i_can_save(const char*file,const char*saveext)
{
	return (saveext&&(strcasecmp(saveext,my_save[0])==0||strcasecmp(saveext,"png")==0))?1:0;
}

int
save_image(const char*file,const char*saveext,const BabyImage*bb,const BabyOptions*bbs)
{
	png_structp			png_ptr;
	png_infop			info_ptr;
	UCHAR				*data, *ptr;
	UINT				x, y;
	png_bytep			row_ptr;
	png_color_8			sig_bit;
	FILE				*f;

	if(file==NULL||saveext==NULL||bb==NULL||bb->num_images<=0||bb->image==NULL)return 0;
	if(strcasecmp(saveext,my_save[0])==0)
		file=my_strdup(file,"png");
	else if(strcasecmp(saveext,"png")==0)
		file=strdup(file);
	else
		return 0;
	if(file==NULL)return 0;

	f = fopen(file, "wb");
	free((void*)file);
	if(f==NULL)return 0;
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL, NULL, NULL);
	if (!png_ptr)
	{
		fclose(f);
		return 0;
	}
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
	{
		fclose(f);
		png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
		return 0;
	}
	if (setjmp(png_ptr->jmpbuf))
	{
		fclose(f);
		png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
		return 0;
	}
	png_init_io(png_ptr, f);
	png_set_IHDR(png_ptr, info_ptr, bb->image->rgb_width, bb->image->rgb_height, 8,
				PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE,
				PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
	sig_bit.red = 8;
	sig_bit.green = 8;
	sig_bit.blue = 8;
	sig_bit.alpha = 8;
	png_set_sBIT(png_ptr, info_ptr, &sig_bit);
	png_write_info(png_ptr, info_ptr);
	png_set_shift(png_ptr, &sig_bit);
	png_set_packing(png_ptr);
	data = (UCHAR*)malloc(bb->image->rgb_width * 4);
	if (!data)
	{
		fclose(f);
		png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
		return 0;
	}
	for (y = 0; y < bb->image->rgb_height; y++)
	{
		ptr = bb->image->rgb_data + (y * bb->image->rgb_width * 3);
		for (x = 0; x < bb->image->rgb_width; x++)
		{
			data[(x << 2) + 0] = *ptr++;
			data[(x << 2) + 1] = *ptr++;
			data[(x << 2) + 2] = *ptr++;
			if (is_mask_color(&bb->image->mask_color)&&
				(data[(x << 2) + 0] == bb->image->mask_color.r) &&
				(data[(x << 2) + 1] == bb->image->mask_color.g) &&
				(data[(x << 2) + 2] == bb->image->mask_color.b))
				data[(x << 2) + 3] = 0;
			else
				data[(x << 2) + 3] = 255;
		}
		row_ptr = data;
		png_write_rows(png_ptr, &row_ptr, 1);
	}
	free(data);
	png_write_end(png_ptr, info_ptr);
	png_destroy_write_struct(&png_ptr, (png_infopp) NULL);
	fclose(f);
	return 1;
}

#ifdef __cplusplus
}
#endif
